# -*- coding: utf-8 -*-
"""
N3 Spatial Codec
================
Updated: 2021-06

Encode an n1 block of data in n3 space using a pseudo hilbert space filling curve

Dependancies
------------
```
from typing import List, Tuple
from scodec.codec.base import SpatialCode
```
Copyright © 2021 LEAP. All Rights Reserved.
"""

from typing import List, Tuple
from scodec.codec.base import SpatialCodec


class N3(SpatialCodec):

    BASE_BLOCK_SIZE = 8   # block size of base iterator

    def __init__(self, block_size: int):
        # check if the block_size is greater than the base block size
        if block_size > self.BASE_BLOCK_SIZE:
            raise NotImplementedError(
                "This version only supports first order\
                (max 8 bit block_size) curves in 3D space"
            )
        super().__init__(block_size, self.BASE_BLOCK_SIZE)
        self.log.info("Configured %s codec with block size: %s", __name__, self.block_size)

    def stream_encode(self, bytestream: bytes, mpl: bool = False) -> List[Tuple[int, int, int]]:
        """
        Encode a stream of bytes in n3 space.

        :param bytestream: block of data for encoding
        :type bytestream: bytes
        :param mpl: flag to enable mpl visualizer, defaults to False
        :type mpl: bool, optional
        :return: encoded stream
        :rtype: List[Tuple[int,int,int]]
        """
        # remove excess bytes if word exceeds an 8 bit number
        bitstream = int(bytestream.hex(), base=16) & 0xFF
        self.log.debug("bitstream: %s", bin(bitstream))
        # unpacked bistream into iterable of single bits
        bits = [bitstream >> i & 0x1 for i in range(self.block_size)]
        # generate index by encoding each set bit sequentially
        stream = list(filter(None, [self.encode(i) if b else None for i, b in enumerate(bits)]))
        self.log.info("stream: %s", stream)
        if mpl: self.render(stream)
        return stream

    def stream_decode(self, stream: List[Tuple[int, int, int]]) -> bytes:
        """
        Decode a stream of coordinates encoded in n3 space into bytes.

        :param stream: stream of n3 space coordinate mapping
        :type stream: List[Tuple[int, int, int]]
        :return: decoded bytestream
        :rtype: bytes
        """
        bitstream = 0x0
        for coor in stream:
            bitstream |= self.decode(coor)
        self.log.info("decoded bitstream: %s", bin(bitstream))
        bytestream = bitstream.to_bytes(self.block_size, byteorder="big", signed=False)
        self.log.info("bytestream: %s", bytestream)
        return bytestream

    def decode(self, coor: Tuple[int, int, int]) -> int:
        """
        Compute bit index from a coordinate tuple encoded from an n3 first order hilbert curve.
        This method simply computes the

        :param coor: n3 space coordinate mapping
        :type coor: Tuple[int, int, int]
        :return: bit index of coordinate
        :rtype: int
        """
        x, y, z = coor
        x, y, z = 1 & x, 1 & y, 1 & z
        d = (x << 2) ^ (x ^ y << 1) ^ (x ^ y ^ z)
        index = 0x1 << d
        self.log.debug("computed index: %s", bin(index))
        return index

    def encode(self, i: int) -> Tuple[int, int, int]:
        """
        Compute coordinate tuple of an n3 hilbert curve at index i. Normally this
        applies iterative mapping to n3 space to constuct hilberts curve @ block_size
        (the iterative solution is not supported therefore the encode function only
        supports the base block_size: 8)

        :param i: bit index
        :type i: int
        :return: coordinate tuple @ bit index i
        :rtype: Tuple[int,int,int]
        """
        # initial coordinates
        x, y, z = self.iterator(i)
        self.log.info("resolved i:%s -> x:%s y:%s z:%s", i, x, y, z)
        return x, y, z

    def iterator(self, i: int) -> Tuple[int, int, int]:
        """
        Base iterator for N3 algorithm

        :param i: probe index
        :type i: int
        :return: base cartesian coordinate of bit @ index i
        :rtype: Tuple[int,int,int]
        """
        r_x = 1 & (i >> 2)
        r_y = 1 & (i >> 1 ^ r_x)
        r_z = 1 & (i ^ r_x ^ r_y)
        return r_x, r_y, r_z

    def transform(
        self, r_x: int, r_y: int, r_z: int, o: Tuple[str, str, str]
    ) -> Tuple[int, int, int]:
        """
        Transform base iterator by applying a reflection about a plane or axis

        :param r_x: x component of iterator coordinate
        :type r_x: int
        :param r_y: y component of iterator coordinate
        :type r_y: int
        :param r_z: z component of iterator coordinate
        :type r_z: int
        :param o: iterator variant selector
        :type o: Tuple[str,str,str]
        :return: transformed coordinate tuple
        :rtype: Tuple[int,int,int]
        """
        # iterator variant selector (-x and x are somehow the same?)
        r_t = r_x, r_y, r_z
        if o[0] == "x":
            if o[1] == "y":
                if o[2] == "z":
                    r_t = r_x, r_y, r_z
                else:
                    r_t = r_x, r_y, 1 - r_z
            elif o[1] == "-y":
                if o[2] == "z":
                    r_t = r_x, 1 - r_y, r_z
                else:
                    r_t = r_x, 1 - r_y, 1 - r_z
            elif o[1] == "z":
                if o[2] == "y":
                    r_t = r_x, r_z, r_y
                else:
                    r_t = r_x, r_z, 1 - r_y
            else:
                if o[2] == "y":
                    r_t = r_x, 1 - r_z, r_y
                else:
                    r_t = r_x, 1 - r_z, 1 - r_y
        elif o[0] == "-x":
            if o[1] == "y":
                if o[2] == "z":
                    r_t = 1 - r_x, r_y, r_z
                else:
                    r_t = 1 - r_x, r_y, 1 - r_z
            elif o[1] == "-y":
                if o[2] == "z":
                    r_t = 1 - r_x, 1 - r_y, r_z
                else:
                    r_t = 1 - r_x, 1 - r_y, 1 - r_z
            elif o[1] == "z":
                if o[2] == "y":
                    r_t = 1 - r_x, r_z, r_y
                else:
                    r_t = 1 - r_x, r_z, 1 - r_y
            else:
                if o[2] == "y":
                    r_t = 1 - r_x, 1 - r_z, r_y
                else:
                    r_t = 1 - r_x, 1 - r_z, 1 - r_y
        elif o[0] == "y":
            if o[1] == "x":
                if o[2] == "z":
                    r_t = r_y, r_x, r_z
                else:
                    r_t = r_y, r_x, 1 - r_z
            elif o[1] == "-x":
                if o[2] == "z":
                    r_t = r_y, 1 - r_x, r_z
                else:
                    r_t = r_y, 1 - r_x, 1 - r_z
            elif o[1] == "z":
                if o[2] == "x":
                    r_t = r_y, r_z, r_x
                else:
                    r_t = r_y, r_z, 1 - r_x
            else:
                if o[2] == "x":
                    r_t = r_y, 1 - r_z, r_x
                else:
                    r_t = r_y, 1 - r_z, 1 - r_x
        elif o[0] == "-y":
            if o[1] == "x":
                if o[2] == "z":
                    r_t = 1 - r_y, r_x, r_z
                else:
                    r_t = 1 - r_y, r_x, 1 - r_z
            elif o[1] == "-x":
                if o[2] == "z":
                    r_t = 1 - r_y, 1 - r_x, r_z
                else:
                    r_t = 1 - r_y, 1 - r_x, 1 - r_z
            elif o[1] == "z":
                if o[2] == "x":
                    r_t = 1 - r_y, r_z, r_x
                else:
                    r_t = 1 - r_y, r_z, 1 - r_x
            else:
                if o[2] == "x":
                    r_t = 1 - r_y, 1 - r_z, r_x
                else:
                    r_t = 1 - r_y, 1 - r_z, 1 - r_x
        elif o[0] == "z":
            if o[1] == "x":
                if o[2] == "y":
                    r_t = r_z, r_x, r_y
                else:
                    r_t = r_z, r_x, 1 - r_y
            elif o[1] == "-x":
                if o[2] == "y":
                    r_t = r_z, 1 - r_x, r_y
                else:
                    r_t = r_z, 1 - r_x, 1 - r_y
            elif o[1] == "y":
                if o[2] == "x":
                    r_t = r_z, r_y, r_x
                else:
                    r_t = r_z, r_y, 1 - r_x
            else:
                if o[2] == "x":
                    r_t = r_z, 1 - r_y, r_x
                else:
                    r_t = r_z, 1 - r_y, 1 - r_x
        elif o[0] == "-z":
            if o[1] == "x":
                if o[2] == "y":
                    r_t = 1 - r_z, r_x, r_y
                else:
                    r_t = 1 - r_z, r_x, 1 - r_y
            elif o[1] == "-x":
                if o[2] == "y":
                    r_t = 1 - r_z, 1 - r_x, r_y
                else:
                    r_t = 1 - r_z, 1 - r_x, 1 - r_y
            elif o[1] == "y":
                if o[2] == "x":
                    r_t = 1 - r_z, r_y, r_x
                else:
                    r_t = 1 - r_z, r_y, 1 - r_x
            else:
                if o[2] == "x":
                    r_t = 1 - r_z, 1 - r_y, r_x
                else:
                    r_t = 1 - r_z, 1 - r_y, 1 - r_x
        return r_t

    def render(self, stream: List[Tuple[int, int, int]]) -> None:
        """
        Render MPL visualizer of index iterator and stream overlay

        :param stream: encoded stream
        :type stream: List[Tuple[int,int,int]]
        """
        index = [self.encode(i) for i in range(self.block_size)]
        self.log.debug("index: %s", index)
        self.log.debug("stream: %s", stream)
        self.visualizer.add_n3_curve(index, marker="", label="index", clr="k")
        self.visualizer.add_n3_curve(stream, marker="o", label="stream", clr="r")
        self.visualizer.show()

    def show_iterators(self) -> None:
        """
        Interactive visualizer for iteration variations
        """
        # variation iterator
        # for curve in Iterators.variations:
        #     order, clr = curve
        #     self.visualizer.add_n3_curve(
        #         d=[self.transform(*self.iterator(i), order) for i in range(8)],
        #         marker='o',
        #         label=str(order),
        #         clr=clr
        #     )
        # self.visualizer.show()



# -*- coding: utf-8 -*-
"""
Spatial Codec™ ABC
==================

Abstract base class implemented by n2 and n3 spatial codec algorithms

Dependancies
------------
```
import math
import logging
from typing import List, Tuple
from abc import ABC, abstractmethod
from scodec.plt.visualizer import Visualizer
```
Copyright © 2021 LEAP. All Rights Reserved.
"""

import math
import logging
from typing import List, Tuple
from abc import ABC, abstractmethod
from scodec.plt.visualizer import Visualizer


class SpatialCodec(ABC):
    def __init__(self, block_size: int, base_block_size: int) -> None:
        self.log = logging.getLogger(__name__)
        # compute codec resolution (next power of 2)
        self.visualizer = Visualizer()
        super().__init__()
        # validate block size
        if not math.log(block_size, base_block_size).is_integer():
            raise ValueError("{} block size must be a power of {}".format(
                __name__, base_block_size))
        self.block_size = block_size
        self._sv = [2**x for x in range(block_size)]
        self.log.debug("s vector: %s", self._sv)

    @abstractmethod
    def stream_encode(self, bytestream: bytes) -> None:
        ...

    @abstractmethod
    def stream_decode(self, coor: List[Tuple], byte_size: int) -> bytes:
        ...

    @abstractmethod
    def decode(self, coor: Tuple) -> int:
        ...

    @abstractmethod
    def encode(self, i: int):
        ...

    @abstractmethod
    def render(self, coors: List[Tuple]):
        ...

    @abstractmethod
    def transform(self):
        ...

    @abstractmethod
    def iterator(self, i: int) -> Tuple:
        ...



import sys
import getopt
import logging

from codec.n3 import N3
from codec.n2 import N2


def main(argv) -> None:
    # defaults
    be = "utf-8"
    dimension = 0
    block = 0
    input_stream = bytes("default", be)
    mpl = False
    # parse opts
    try:
        opts, _ = getopt.getopt(
            argv, "n:b:d:v:", ["dimension=", "block=", "data=", "verbose="])
    except getopt.GetoptError:
        logging.exception("python -m sc -n 2 -b 32 -s test -v=")
        sys.exit(2)
    for opt, arg in opts:
        if opt in ("-d", "--data"):
            input_stream = bytes(arg, be)
        elif opt in ("-n", "--dimension"):
            dimension = int(arg)
        elif opt in ("-v, --verbose"):
            mpl = True
        elif opt in ("-b, --block"):
            block = int(arg)
    logging.info("Input stream: %s", input_stream)
    logging.info("Byte Encoding: %s", be)
    logging.info("Block size: %s", block)
    logging.info("Encoding dimension: %s", dimension)
    logging.info("MPL Visualizer: %s", mpl)
    # N2/N3 impl split
    if dimension == 2:
        n2_sc = N2(block)
        encode_stream = n2_sc.stream_encode(input_stream, mpl=mpl)
        bytestream = n2_sc.stream_decode(encode_stream, len(input_stream))
    elif dimension == 3:
        n3_sc = N3(block)
        encode_stream = n3_sc.stream_encode(input_stream, mpl=mpl)
        bytestream = n3_sc.stream_decode(encode_stream, len(input_stream))
    else:
        raise ValueError("Spatial codec is only defined for 2D and 3D space filling curves")
    print(bytestream.decode("utf-8"))


if __name__ == "__main__":
    logger = logging.getLogger(__name__)
    main(sys.argv[1:])



"""
Spatial Codec __init___
===================================

Initialization for spatial codec

Copyright © 2021 LEAP. All Rights Reserved.
"""

import yaml
from pathlib import Path
import logging
import logging.config

from scodec.__version__ import __version__


# cleanup all previous logs for new runtime environment
CONFIG_PATH = Path(__file__).parent.joinpath("config/log.yaml")

# check for existance of config.yaml
if not CONFIG_PATH.exists(): raise FileNotFoundError

# configure the logger
with open(CONFIG_PATH) as file:
    logging.config.dictConfig(yaml.full_load(file))

_log = logging.getLogger(__name__)
_log.info("scodec version: %s", __version__)



# -*- coding: utf-8 -*-
"""
N2 Spatial Codec
================
Updated: 2021-06

Encode an n1 block of data in n2 space using a pseudo hilbert space filling curve

Dependancies
------------
```
from typing import List, Tuple
from scodec.codec.base import SpatialCodec
```
Copyright © 2021 LEAP. All Rights Reserved.
"""

from typing import List, Tuple
from scodec.codec.base import SpatialCodec


class N2(SpatialCodec):

    BASE_BLOCK_SIZE = 4   # block size of base iterator

    def __init__(self, block_size: int):
        super().__init__(block_size, self.BASE_BLOCK_SIZE)
        self.log.info("Configured %s codec with block size: %s", __name__, self.block_size)

    def stream_encode(self, bytestream: bytes, mpl: bool = False) -> List[Tuple[int, int]]:
        """
        Encode a stream of bytes in n2 space.

        :param bytestream: block of data for encoding
        :type bytestream: bytes
        :param mpl: flag to enable mpl visualizer, defaults to False
        :type mpl: bool, optional
        :return: encoded stream
        :rtype: List[Tuple[int,int]]
        """
        # remove excess bytes if word exceeds resolution
        bitstream = int(bytestream.hex(), base=16) & (2 ** self.block_size - 1)
        self.log.debug("bitstream: %s", bin(bitstream))
        bits = [bitstream >> i & 0x1 for i in range(self.block_size)]
        index = list(filter(None, [self.encode(i) if b else None for i, b in enumerate(bits)]))
        self.log.info("index: %s", index)
        if mpl: self.render(index)
        return index

    def stream_decode(self, stream: List[Tuple[int, int]]) -> bytes:
        """
        Decode a stream of coordinates encoded in n2 space into bytes.

        :param stream: stream of n2 space coordinate mapping
        :type stream: List[Tuple[int,int]]
        :return: decoded bytestream
        :rtype: bytes
        """
        bitstream = 0x0
        for coor in stream:
            bitstream |= self.decode(coor)
        self.log.info("decoded bitstream: %s", bin(bitstream))
        bytestream = bitstream.to_bytes(self.block_size, byteorder="big", signed=False)
        self.log.info("bytestream: %s", bytestream)
        return bytestream

    def decode(self, coor: Tuple[int, int]) -> int:
        """
        Compute bit index from a coordinate tuple encoded from an n2 hilbert curve.
        This method iteratively applies the inverse operation of the n2 space encoding.

        :param coor: n2 space coordinate mapping
        :type coor: Tuple[int,int]
        :return: bit index of coor
        :rtype: int
        """
        d = 0
        s = self.block_size >> 1
        x, y = coor
        self.log.debug("n: %s x: %s y: %s", self.block_size, x, y)
        while s > 0:
            rx = (x & s) > 0
            ry = (y & s) > 0
            d += s ** 2 * ((3 * rx) ^ ry)
            x, y = self.transform(x, y, rx, ry, self.block_size)
            self.log.debug("s: %s rx: %s ry:%s x:%s y:%s", s, rx, ry, x, y)
            s = s >> 1  # divide by 2 each iteration
        self.log.debug("d: %s x: %s y: %s", d, x, y)
        index = 0x1 << d
        self.log.debug("computed index: %s", bin(index))
        return index

    def encode(self, i: int) -> Tuple[int, int]:
        """
        Compute coordinate tuple of an n2 hilbert curve at index i. This applies
        iterative mapping to n2 space to constuct hilberts curve @ resolution

        :param i: bit index
        :type i: int
        :return: coordinate tuple @ bit index i
        :rtype: Tuple[int,int]
        """
        index = i
        self.log.debug("Computing coordinate at bit: %s", i)
        # initial coordinates
        x, y = 0, 0
        for level, c in enumerate(self._sv):
            # Once the index reaches 0 the x and y bits are latched and alternate between each other
            # before converging before we exceed the range boundary n
            if i == 0:
                # we are done
                if x == y:
                    break
                # compute last flip (i required for forcasting)
                x, y = (y, x) if (self.block_size - level) & 1 else (x, y)
                break
            # generate base iterator
            r_x, r_y = self.iterator(i)
            x, y = self.transform(x, y, r_x, r_y, c)
            # translate base iterator
            x += c * r_x  # x = x + (s or 0)
            y += c * r_y  # y = y + (s or 0)
            i = i >> 2  # regions of seperation (4 verticies)
            self.log.debug(
                "i:%s cells:%s | i/2 odd:%s i/2 odd and i odd:%s -> x:%s y:%s",
                i,
                c,
                r_x,
                r_y,
                x,
                y,
            )
        self.log.debug("resolved i:%s -> x:%s y:%s", index, x, y)
        return x, y

    def transform(self, x: int, y: int, r_x: int, r_y: int, c: int) -> Tuple[int, int]:
        """
        Transform base iterator by applying a reflection about an axis or line by cell selector c

        :param x: x translation to cell index c
        :type x: int
        :param y: y translation to cell index c
        :type y: int
        :param r_x: x component of base iterator coordinate
        :type r_x: int
        :param r_y: y component of base iterator coordinate
        :type r_y: int
        :param s: cell index
        :type s: int
        :return: transformed coordinate tuple from cell c
        :rtype: Tuple[int,int]
        """
        # rotation function for this region
        if r_y == 0:
            if r_x == 1:
                x, y = c - 1 - x, c - 1 - y
            x, y = y, x
        return x, y

    def iterator(self, i: int) -> Tuple[int, int]:
        """
        Base iterator for N2 algorithm

        :param i: probe index
        :type i: int
        :return: base cartesian coordinate of bit @ index i
        :rtype: Tuple[int,int]
        """
        r_x = 1 & (i >> 1)
        r_y = 1 & (i ^ r_x)
        return r_x, r_y

    def render(self, stream: List[Tuple[int, int]]) -> None:
        """
        Render MPL visualizer of index iterator and stream overlay

        :param stream: encoded stream
        :type stream: List[Tuple[int,int]]
        """
        index = [self.encode(i) for i in range(self.block_size)]
        self.log.debug("index: %s", index)
        self.log.debug("stream: %s", stream)
        self.visualizer.add_n2_curve(index, marker="", label="index", clr="k")
        self.visualizer.add_n2_curve(stream, marker="o", label="stream", clr="r")
        self.visualizer.show()



# -*- coding: utf-8 -*-
"""
Visualizer
==========
Contributors: Christian Sargusingh
Updated: 2021-05

Interactive spatial codec rendering using mpl api.

Dependancies
------------
```
from typing import List, Tuple
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.widgets import CheckButtons
```
Copyright © 2021 LEAP. All Rights Reserved.
"""

from typing import List, Tuple
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D  # noqa: F401 unused import
from matplotlib.widgets import CheckButtons


class Visualizer:
    def __init__(self) -> None:
        plt.style.use("dark_background")
        self.fig = plt.figure()
        self.ax = self.fig.add_subplot(111, projection="3d")
        self.curves = []
        self.cube_max = 0

    def add_n3_curve(
        self, d: List[Tuple[int, int, int]], marker: str, label: str, clr: str
    ) -> None:
        x = list(map(lambda x: x[0], d))
        y = list(map(lambda x: x[1], d))
        z = list(map(lambda x: x[2], d))
        # Update cube max
        self.cube_max = max(max(x), max(y), max(z), self.cube_max)
        # Don't mess with the limits!
        (l,) = self.ax.plot(x, y, z, visible=True, marker=marker, color=clr, label=label)
        self.curves.append(l)

    def add_n2_curve(self, d: List[Tuple[int, int]], marker: str, label: str, clr: str) -> None:
        x = list(map(lambda x: x[0], d))
        y = list(map(lambda x: x[1], d))
        # Update cube max
        self.cube_max = max(max(x), max(y), self.cube_max)
        # Don't mess with the limits!
        (l,) = self.ax.plot(x, y, visible=True, marker=marker, color=clr, label=label)
        self.curves.append(l)

    def show(self) -> None:
        def func(label):
            index = labels.index(label)
            self.curves[index].set_visible(not self.curves[index].get_visible())
            plt.draw()

        # Make checkbuttons with all plotted lines with correct visibility
        rax = plt.axes([0.05, 0.1, 0.2, 0.7])
        labels = [str(curve.get_label()) for curve in self.curves]
        visibility = [bool(curve.get_visible()) for curve in self.curves]
        check = CheckButtons(rax, labels, visibility)
        self.ax.set_xlim(0, self.cube_max)
        self.ax.set_ylim(0, self.cube_max)
        plt.autoscale(False)
        self.ax.set_xlabel("X")
        self.ax.set_ylabel("Y")
        self.ax.set_zlabel("Z")
        self.ax.set_zlim(0, self.cube_max)
        plt.subplots_adjust(left=0.2)
        check.on_clicked(func)
        plt.show()
